#pragma once

#include <Base\Math\quat.h>
#include <Base\Math\vec3.h>
#include <Base\Math\mat4.h>
#include <Base\String\String.hpp>

#include <string_view>
#include <vector>
#include <optional>
#include <memory>
#include <variant>

namespace hyplex
{
	struct BasicBufferedLexState;
}

namespace hyplex::drakan
{
	

#pragma pack(push)
#pragma pack(1)
	struct Header
	{
		char id[4];
		uint16_t version;
		uint32_t directoryOffset;
		uint16_t numEntries;
	};

	struct Entry
	{
		uint8_t type;
		uint8_t subtype;
		uint16_t unk00;
		uint16_t parentDirectory;
		uint32_t offset;
		uint32_t size;
	};
#pragma pack(pop)

	enum class EDataType: uint16_t;
	struct Database;

	struct Dependency
	{
		unsigned int id = 0;
		std::string path;
	};

	struct DatabaseCollection
	{
		std::string dir;
		std::string name;
		std::vector<Dependency> dependencies;
	};

	enum class ENodeType
	{
		DIRECTORY,
		MODEL,
		SOUND,
		TEXTURE,
		ANIMATION,
		OBJECT,
		SEQUENCE,
		LEVEL_METADATA,
		LEVEL_TERRAIN,
		LEVEL_LAYER_GROUPING,
		LEVEL_ENTITY_COLLECTION,
	};

	struct NodeDumpArg
	{
		const DatabaseCollection* dbMeta = nullptr;
		const Database* db = nullptr;
		string_view outputRootDir;
		string_view dbRootDir;
		string_view dir;
	};

	struct INode
	{
		virtual ENodeType getType() const noexcept = 0;
		virtual void dumpNode(const NodeDumpArg& arg) const;
		virtual ~INode() = default;
	};

	struct Directory : INode
	{
		String name;
		std::vector<std::unique_ptr<INode>> children;

		virtual ENodeType getType() const noexcept override;
		virtual void dumpNode(const NodeDumpArg& arg) const override;
	};

	struct Model : INode
	{
		struct Polygon
		{
			uint16_t n; // 3 or 4
			uint16_t unk;
			uint16_t texId;
			std::array<uint16_t, 4> indices;
			std::array<fvec2, 4> uvs;
		};

		struct VertexWeight
		{
			unsigned int vertexI;
			float weight;
		};

		struct BoundingInfo
		{
			// Multiply this stuff by 2048 to get the editor's units.
			fvec3 boundingSphereCenter{}; // probably
			float boundingSphereRadius = 0;
			float unk0[3]{};
			float boundingBoxSizeX = 0;
			float unk1[3]{};
			float boundingBoxSizeY = 0;
			float unk2[3]{};
			float boundingBoxSizeZ = 0;
		};

		struct SceneNode
		{
			String name;
			int meshI = -1;
			int childI = -1;
			int siblingI = -1;
			fmat3x4 worldXForm = identity;
			std::vector<VertexWeight> vertexWeights;
		};

		struct LODInfo
		{
			struct Mesh
			{
				uint16_t meshI;
				uint16_t unk;
				uint32_t usage;
				uint32_t nodeI;
				uint32_t firstVertexI;
				uint32_t numVertices;
				uint32_t firstPolyI;
				uint32_t numPolys;
			};

			String name;
			std::vector<Mesh> meshes;
		};

		struct Channel
		{
			struct RawLODCapInfo
			{
				uint32_t firstCapPolyI;
				uint32_t numCapPolys;
				uint32_t firstPartPolyI;
				uint32_t numPartPolys;
				uint32_t unk;
			};
			struct LODCapInfo: RawLODCapInfo
			{
				std::vector<VertexWeight> weights;
			};

			unsigned int nodeI = 0;
			fmat3x4 xformA = identity;
			fmat3x4 xformB = identity;
			std::vector<LODCapInfo> lods;
		};

		struct Sphere
		{
			uint16_t unk00;
			uint16_t otherSphereI;
			uint32_t vertexI;
			float radius;
			uint32_t unk0C;
		};

		String name;
		unsigned int id = 0;
		unsigned int flags = 0;
		std::vector<fvec3> vertices;
		std::vector<unsigned int> textureIds;
		std::vector<Polygon> polys;

		BoundingInfo boundingInfo{};
		std::vector<SceneNode> nodes;
		std::vector<LODInfo> lods;
		std::vector<uint32_t> animIds;
		std::vector<Channel> channels;
		std::vector<Sphere> boundingSpheres;

		void consumeData(const Entry& entry, BasicBufferedLexState& bin);
		String buildCollada(const NodeDumpArg& arg) const;
		virtual ENodeType getType() const noexcept override;
		virtual void dumpNode(const NodeDumpArg& arg) const override;
	};

	struct Sound: INode
	{
		unsigned int id = 0;
		String name;
		unsigned int unk = 0;
		unsigned int numChannels = 0;
		unsigned int bitsPerSample = 0;
		unsigned int sampleRate = 0;
		int volume = 0;
		float dropOffDistance = 0;
		unsigned int priority = 0;
		std::vector<char> blob;

		virtual ENodeType getType() const noexcept override;
		virtual void dumpNode(const NodeDumpArg& arg) const override;
	};

	struct Texture: INode
	{
		unsigned int id = 0;
		unsigned int width = 0;
		unsigned int height = 0;
		unsigned int pitch = 0;
		unsigned int alphaBitsPerPixel = 0;
		unsigned int unk5 = 0;
		unsigned int unk6 = 0;
		unsigned int unk7 = 0;
		unsigned int bumpMapTexId = 0;
		unsigned short bitsPerPixel = 0;
		uint8_t animationStartFrameFPS = 0;
		bool isAnimationNextFrame = false;
		unsigned int materialObjectId = 0;
		unsigned int unk11 = 0;
		unsigned int unk12 = 0;

		std::vector<char> blob;

		void consumeData(const Entry& entry, BasicBufferedLexState& bin);
		virtual ENodeType getType() const noexcept override;
		virtual void dumpNode(const NodeDumpArg& arg) const override;

		static void dumpAnimationInfo(const Directory& textureDirectory, string_view outputDir);
	};

	struct Animation : INode
	{
		struct NodeFramesIndexRange
		{
			unsigned int firstFrameI = 0;
			unsigned int numFrames = 0;
		};

		struct Frame
		{
			float t = 0;
			fmat3x4 xform = identity;
		};

		unsigned int id = 0;
		String name;
		float duration = 0;
		unsigned int originalNumFrames = 0;
		unsigned int numFrames = 0;
		unsigned int unk0 = 0;
		unsigned int modelNumNodes = 0;
		unsigned int modelNumChannels = 0;
		unsigned int unk1 = 0; // Non-Looping flag?
		unsigned int refCount = 0;
		float rotationThreshold = 0;
		float scaleThreshold = 0;
		float translationThreshold = 0; // Probably.
		std::vector<NodeFramesIndexRange> nodeFramesIndexRanges;
		std::vector<Frame> frames;

		void consumeData(const Entry& entry, BasicBufferedLexState& bin);
		virtual ENodeType getType() const noexcept override;
		virtual void dumpNode(const NodeDumpArg& arg) const override;
	};

	
	

	struct Database
	{
		std::vector<char> blob;
		std::vector<Entry> entries;

		Directory rootDir;
		std::vector<std::array<uint8_t, 4>> palette;

		void parse();

		String getEntriesDump() const;
	};

	String trimNul(String str);

	String escapeToFilename(string_view s);

	std::optional<Database> loadDatabase(std::string_view inputPath);
}