#include "Object.h"
#include "RFL.h"

#include <Base\Containers\range_for_std_containers.hpp>
#include <Base\IO\FileUtil.hpp>
#include <Base\IO\MemoryReadStream.h>
#include <Base\IO\ReadStreamBinLex.h>
#include <Base\IO\ReadStreamBinLex_stdvector.h>
#include <Base\Math\math_strfmt.h>
#include <Base\String\Formatting.h>

namespace hyplex::drakan
{
	static const bool kEnableDump = true;

	std::vector<Object::Property> Object::readProperties(BasicBufferedLexState& bin, bool hasIndices)
	{
		uint32_t numProps = read<uint32_t>(bin);
		uint32_t numPropDWords = read<uint32_t>(bin);

		if (hasIndices)
			std::swap(numProps, numPropDWords);

		std::vector<Property> props;
		props.resize(numProps);

		const BasicBufferedLexState originalDwords{ consumeUptoN(bin, numPropDWords * 4) };

		if (hasIndices)
		{
			for (unsigned int i = 0; i < numProps; ++i)
				props[i].index = read<uint16_t>(bin);
		}

		BasicBufferedLexState dwords = originalDwords;

		size_t totalListDWords = 0;

		for (unsigned int i = 0; i < numProps; ++i)
		{
			const DataType type(read<uint32_t>(bin));

			Property& prop = props[i];
			prop.type = type.elementType;
			prop.name = trimNul(readString(bin, read<uint16_t>(bin)));
			switch (prop.type)
			{
			case EDataType::INT:
			case EDataType::CHOICE: // By default, this is a boolean, but an enum string list may be supplied.
				HYPLEX_ENFORCE(!type.isList);
				prop.value = read<int32_t>(dwords);
				break;
			case EDataType::ANIM_CHANNEL:
				if(!type.isList)
					prop.value = read<int32_t>(dwords);
				else
				{
					const unsigned int base = read<uint16_t>(dwords);
					const unsigned int count = read<uint16_t>(dwords);
					BasicBufferedLexState listBin{ originalDwords.buf.slice_n(base * 4, count * 4) };
					std::vector<int> list;
					list.reserve(count);
					for (unsigned int j = 0; j < count; ++j)
						list.push_back(read<int32_t>(listBin));
					totalListDWords += count;
					prop.value = std::move(list);
				}
				break;
			case EDataType::FLOAT:
				if(!type.isList)
					prop.value = read<float>(dwords);
				else
				{
					const unsigned int base = read<uint16_t>(dwords);
					const unsigned int count = read<uint16_t>(dwords);
					BasicBufferedLexState listBin{ originalDwords.buf.slice_n(base * 4, count * 4) };
					std::vector<float> list;
					list.reserve(count);
					for (unsigned int j = 0; j < count; ++j)
						list.push_back(read<float>(listBin));
					totalListDWords += count;
					prop.value = std::move(list);
				}
				break;
			case EDataType::OBJECT:
			case EDataType::MODEL:
			case EDataType::SOUND:
			case EDataType::ANIMATION:
			case EDataType::SEQUENCE:
			case EDataType::TEXTURE:
			case EDataType::LANDSCAPE_LAYER: // Not really an object.
				if(!type.isList)
					prop.value = ObjectRef{ read<uint32_t>(dwords) };
				else
				{
					const unsigned int base = read<uint16_t>(dwords);
					const unsigned int count = read<uint16_t>(dwords);
					BasicBufferedLexState listBin{ originalDwords.buf.slice_n(base * 4, count * 4) };
					std::vector<ObjectRef> list;
					list.reserve(count);
					for (unsigned int j = 0; j < count; ++j)
						list.push_back(ObjectRef{ read<uint32_t>(listBin) });
					totalListDWords += count;
					prop.value = std::move(list);
				}
				break;
			case EDataType::COLOUR:
			{
				HYPLEX_ENFORCE(!type.isList);
				const Colour bgra = read<Colour>(dwords);
				prop.value = Colour{ bgra[2], bgra[1], bgra[0], bgra[3] };
				break;
			}

			case EDataType::STRING:
			{
				HYPLEX_ENFORCE(!type.isList);
				const unsigned int base = read<uint16_t>(dwords);
				const unsigned int count = read<uint16_t>(dwords);
				BasicBufferedLexState listBin{ originalDwords.buf.slice_n(base * 4, count * 4) };
				prop.value = consumeUntilNul(listBin);
				totalListDWords += count;
				break;
			}

			case EDataType::UNK6:


			case EDataType::UNK13:
				HYPLEX_ENFORCE_UNREACHABLE();
			}
		}

		// Fails in "Multiplayer\Monastery\Junction DB.odb". Unused string.
		//HYPLEX_ENFORCE(dwords.buf.size() == totalListDWords * 4);

		return props;
	}

	void Object::consumeData(const Entry& entry, BasicBufferedLexState& bin)
	{
		(void)(entry, bin);

		if (entry.type == 2 && entry.subtype == 1)
		{
#pragma pack(push)
#pragma pack(1)
			struct Header
			{
				uint16_t unk00;
				uint32_t modelId;
				uint16_t classId;
				uint16_t unk08;
			};
#pragma pack(pop)

			name = trimNul(readString(bin, read<uint16_t>(bin)));
			const Header header = read<Header>(bin);

			modelId = header.modelId;
			classId = header.classId;
			props = readProperties(bin);
		}
	}
	ENodeType Object::getType() const noexcept
	{
		return ENodeType::OBJECT;
	}

	String Object::Property::toString() const
	{
		String str;

		struct ValueDumper
		{
			String& str;

			void operator()(std::monostate)
			{
				str += "[no value]";
			}

			void operator()(int x)
			{
				printfmt(str, "%i", x);
			}

			void operator()(float x)
			{
				printfmt(str, "%f", x);
			}

			void operator()(string_view x)
			{
				printfmt(str, "\"%s\"", x);
			}

			void operator()(ObjectRef x)
			{
				printfmt(str, "object %04x", x.id);
			}

			void operator()(Colour x)
			{
				printfmt(str, "colour %u %u %u %u", x[0], x[1], x[2], x[3]);
			}

			void operator()(const ObjectRefList& x)
			{
				str += "objects"_sv;
				for (const ObjectRef obj : x)
					printfmt(str, " %04x", obj.id);
			}

			void operator()(const std::vector<int>& x)
			{
				printfmt(str, "%i", asArrayView(x));
			}

			void operator()(const std::vector<float>& x)
			{
				printfmt(str, "floats %f", asArrayView(x));
			}

		} valueDumper{ str };

		str += name;
		str += " = "_sv;

		std::visit(valueDumper, value);

		return valueDumper.str;
	}

	void Object::dumpNode(const NodeDumpArg& arg) const
	{
		if (!kEnableDump)
			return;

		String str;

		printfmt(str, "id %u\n", id);
		printfmt(str, "name %s\n", name);
		printfmt(str, "model %u\n", modelId);
		printfmt(str, "class %02x\n", classId);

		str += '\n';

		for (const Property& prop : props)
		{
			str += prop.toString();
			str += '\n';
		}

		FileUnbufferedWriteStream(sprintfmt("%sObject%u %s.txt", arg.dir, id, escapeToFilename(name)), EFileWriteOpenMode::RECREATE, true)
			.write(asRawUChars(asArrayView(str)));
	}
}